/*
 * File lexical.c
 * Lexically scanning BLISS
 *
 * The Unknown Scandinavian was here
 *
 * Spec		see lexical.h
 * Test files	t2.b, t3.b, t1.b, psint.b
 *
 * Known deficiencies:
 *	You can only scan one BLISS file at one time.
 */

#include <stdio.h>
#include <ctype.h>
#include "lexical.h"

/*********************************************************************
 * Externally available data
 */

int	token_int;			/* value of digits */
char	token_string[256];		/* names and quoted-strings */
int	token_dollars;			/* dollars in a name */
int	token_overflow;
int	token_collision;		/* tokens collide */

/* Where the token starts: */

int	token_byte;			/* # of chars scanned */
int	token_line;			/* # of lines scanned */
int	token_col;			/* # of columns of last line */
int	token_count;			/* # of tokens scanned */

/*********************************************************************
 * Static data and forward function declarations
 */

static int t_byte;			/* The scanning point */
static int t_line;
static int t_col;
static int t_count;
static FILE *scan = (FILE *) NULL;

static int popchar ();
static int pushchar ();
     
/*********************************************************************
 * Functions
 */
     
int newfile (name)
     char *name;
{
  t_byte = 0;
  t_line = 1;
  t_col  = 0;
  t_count = 0;
  
  if (scan != (FILE *) NULL)
    fclose (scan);
  scan = fopen (name, "r");
  if (scan != (FILE *) NULL)
    return 0;
  else
    return -1;
}

int nofile ()
{
  int i;
  
  i = fclose (scan);
  scan = (FILE *) NULL;
  return i;
}

int token ()				/* For compilers */
{
  int t;
  
  while (TOKEN_EOF != (t = raw_token ())) 
    {
      if (t != TOKEN_ECOMM && t != TOKEN_TCOMM)
	break;
    }
  return t;
}

int raw_token ()			/* If you want comments */
{
  int c;
  
  token_string[0] = '\0';
  token_int = 0;
  token_dollars = 0;
  token_overflow = 0;
  token_collision = 0;
  
  while (isspace (c = popchar ()))
    ;
  if (c == EOF)
    return TOKEN_EOF;
  
  token_count ++;
  token_byte = t_byte;
  token_line = t_line;
  token_col  = t_col;
  
  if (isalpha (c))
    return name (c);
  if (isdigit (c))
    return digits (c);
  switch (c)
    {
    case '_':
    case '$':
      return name (c);
    case '%':				/* percent or e-comm */
      c = popchar ();
      if (c != '(')
	{
	  pushchar (c);
	  return TOKEN_PERC;
	}
      else
	{				/* Embedded comment */
	  while (EOF != (c = popchar ()))
	    {
	      token_int ++;
	      if (c == ')')
		{
		  c = popchar ();
		  if (c == '%')
		    {
		      token_int --;
		      return TOKEN_ECOMM;
		    }
		  else
		    {
		      pushchar (c);	/* wrong ')' */
		    }
		}
	    }
	  token_collision = 1;
	  return TOKEN_ECOMM;
	}
      /* NOTREACHED */
      break;
    case '!':				/* t-comm */
      while (EOF != (c = popchar ()))
	{
	  if (c == '\n')
	    break;
	  token_int ++;			/* length of comment */
	}
      return TOKEN_TCOMM;
    case '\'':				/* quoted string */
      while (EOF != (c = popchar ()))
	{
	  if (c == '\'')
	    {
	      c = popchar ();
	      if (c == '\'')
		{
		  if (token_int < 255)
		    token_string[token_int] = (char) c;
		  else
		    token_overflow = 1;
		  token_int ++;
		}
	      else
		{
		  pushchar (c);
		  token_string[token_int<255? token_int: 255] = '\0';
		  return TOKEN_STRING;
		}
	    }
	  else				/* ordinary char */
	    {
	      if (iscntrl(c) && c != '\t')
		{
		  c = ' ';		/* safe for zero */
		  token_collision = 1;
		}
	      if (token_int < 255)
		token_string[token_int] = (char) c;
	      else
		token_overflow = 1;
	      token_int ++;
	    }
	}
      token_collision = 1;		/* unexpected EOF */
      token_string[token_int<255 ? token_int : 255] = '\0';
      return TOKEN_STRING;
    case '.':
      return TOKEN_DOT;
    case '^':
      return TOKEN_POW;
    case '*':
      return TOKEN_MUL;
    case '/':
      return TOKEN_DIV;
    case '+':
      return TOKEN_PLUS;
    case '-':
      return TOKEN_MINUS;
    case '=':
      return TOKEN_ASSIGN;
    case ',':
      return TOKEN_COMMA;
    case ';':
      return TOKEN_SEMI;
    case ':':
      return TOKEN_COLON;
    case '(':
      return TOKEN_LPAREN;
    case ')':
      return TOKEN_RPAREN;
    case '[':
      return TOKEN_LBRACK;
    case ']':
      return TOKEN_RBRACK;
    case '<':
      return TOKEN_LANG;
    case '>':
      return TOKEN_RANG;
    default:
      return TOKEN_NIL;
    }
}

int name (c)
     int c;
{
  
  if (c == '$')
    token_dollars = 1;
  token_string[0] = (char) isupper(c) ? tolower(c) : c;
  token_int = 1;
  
  while (EOF != (c = popchar ()))
    {
      if (isalpha (c) || c == '_' || c == '$' || isdigit (c))
	{
	  if (token_int<31)
	    token_string[token_int] =
	      (char) isupper(c) ? tolower(c) : c;
	  else
	    token_overflow = 1;
	  token_int ++;
	  if (c == '$')			/* For DEC names only */
	    token_dollars = 1;
	}
      else
	{
	  if (c == '%')
	    token_collision = 1;
	  pushchar (c);
	  break;
	}
    }
  token_string[token_int<31?token_int:31] = '\0';
  return TOKEN_NAME;
}

int digits (c)
     int c;
{
  token_collision = 0;
  token_int = c - '0';
  
  while (isdigit (c = popchar ()))
    {
      token_int *= 10;
      token_int += c - '0';
    }
  if (isalpha (c) || c == '_' || c == '$')
    token_collision = 1;
  pushchar (c);
  return TOKEN_DIGITS;
}

static int popchar ()
{
  int c;
  
  c = getc (scan);
  if (c != EOF)
    {
      t_byte ++;
      t_col ++;
    }
  if (c == '\f')
    c = '\n';
  if (c == '\n')
    {
      t_line ++;
      t_col = 0;
    }
  return c;
}

static int pushchar (c)
     int c;
{
  if (c != EOF)
    {
      t_byte --;
      t_col --;
    }
  if (c == '\n')
    t_line --;
  /* Well, if we push a newline, we will soon pop it again
   * and thus we don't have to care about t_col here.
   */
  return ungetc (c, scan);
}
